static char *sccsid = "@(#) SupportPacks/ms0c/public/mqwrap.c, supportpacks, p000, p000-L000330          1.2  00/03/06 17:46:43";

/**********************************************************************/
/*                                                                    */
/* Module Name: MQWRAP.C                                              */
/*    A wrapper for the Entrust-based channel exits, needed for AIX   */
/*                                                                    */
/* Description: Wrap the C++ libraries into something usable without  */
/*    a C++ environment                                               */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Function:                                                          */
/*   On AIX, MQSeries channel exits cannot be written in C++. This    */
/*   is because the code internally uses the load() function. The     */
/*   Entrust runtime support requires a C++ environment, which is     */
/*   not available. So, to solve this, we can write a wrapper module  */
/*   which appears to the MCAs to be a channel exit but which         */
/*   gets the C++ environment installed before calling the real       */
/*   exit. This is done by using the loadAndInit() function - which   */
/*   appears identical to load(), but is precisely what we need here. */
/*                                                                    */
/*   We then passthrough the parameters to this exit, calling the     */
/*   real module directly. Rather than worry about function names etc */
/*   within that module, I've defined MQStart (which is normally      */
/*   a dummy function) to be not only the entrypoint returned by      */
/*   loadAndInit but also the single 'external' function used for the */
/*   actual work. This makes the wrapper's job easier as it only has  */
/*   to remember the EP value and call it.                            */
/*                                                                    */
/*   There is a corresponding terminateAndUnload, which I've          */
/*   commented out for now because it appeared to crash on this       */
/*   version of AIX. It shouldn't really matter too much for now, but */
/*   if I ever work out why (probably an AIX APAR is needed) then     */
/*   I'll reinstate it.                                               */
/*                                                                    */
/**********************************************************************/
/*                                                                    */
/* Notes:                                                             */
/*   Language:     C                                                  */
/*     Howver, in order to build and run this module, you must have   */
/*     the C++ runtime (xlC.rte) installed.                           */
/**********************************************************************/

#include <pthread.h>
#include <stdio.h>
#include <sys/ldr.h>

#include <cmqc.h>
#include <cmqxc.h>

pthread_mutex_t m;
pthread_once_t onceblock = PTHREAD_ONCE_INIT;

static void onceonly(void);

static int loadcount = 0;
static int mutex_avail = 0;
static int (*ep)() = 0;

/**********************************************************************/
/* These defs are copied from the ldr.h file - I couldn't find the    */
/* correct header file for loadAndInit                                */
/**********************************************************************/
typedef int LOAD_T ();
extern LOAD_T *loadAndInit(char *, uint, char *);

/**********************************************************************/
/* This is the real exit module - it ought to be a parameter or ini   */
/* value, but I've hardcoded it here for ease of programming! It      */
/* should always be a fully-qualified pathname. Don't rely on         */
/* LIBPATH and friends to do the work for you.                        */
/**********************************************************************/
#define MQSECCE "/var/mqm/exits/libmqsecces"

MQStart(
    PMQVOID  pChannelExitParms,   /* Channel exit parameter block */
    PMQVOID  pChannelDefinition,  /* Channel definition           */
    PMQLONG  pDataLength,         /* Length of data               */
    PMQLONG  pAgentBufferLength,  /* Length of agent buffer       */
    PMQVOID  pAgentBuffer,        /* Agent buffer                 */
    PMQLONG  pExitBufferLength,   /* Length of exit buffer        */
    PMQPTR   pExitBufferAddr)     /* Address of exit buffer       */
{
  PMQCXP        pParms;
  FILE *fp;
  pParms   = (PMQCXP)pChannelExitParms;

#ifdef DEBUG
  fp = fopen("/tmp/mqsecche.log","a");
  setbuf(fp,NULL);
  fprintf(fp,"mqwrap called\n");
#endif

  /***************************************************************************/
  /* Take a process-wide lock while we load up the real exit code.           */
  /* This will let us be sure that the loadcount is correct, in case         */
  /* the channel initiator decides to start two channels simultaneously.     */
  /***************************************************************************/
  pthread_once(&onceblock,onceonly);
  pthread_mutex_lock(&m);

  if (pParms->ExitReason == MQXR_INIT)
  {
    if (loadcount==0)
    {
#ifdef DEBUG
      fprintf(fp,"About to load %s\n",MQSECCE);
#endif
      ep = loadAndInit(MQSECCE,0,NULL);
#ifdef DEBUG
      fprintf(fp,"Loaded EP = %08X\n",ep);
#endif
    }
    else
    {
#ifdef DEBUG
      fprintf(fp,"INIT called, already loaded\n");
#endif
    }
    if (ep)
      loadcount++;
  }

  /***************************************************************************/
  /* This is where the real work is done. Once the main module has been      */
  /* loaded, its entry point (MQStart) is invoked directly with the          */
  /* parameters being passed straight through.                               */
  /***************************************************************************/
  if (ep)
  {
#ifdef DEBUG
     fprintf(fp,"About to call real exit - loadcount = %d\n",loadcount);
#endif
     ((MQCHANNELEXIT *)ep)(pChannelExitParms, pChannelDefinition, pDataLength,
	pAgentBufferLength, pAgentBuffer, pExitBufferLength,
	pExitBufferAddr);
#ifdef DEBUG
     fprintf(fp,"Returned from real exit\n");
#endif
  }

  if (pParms->ExitReason == MQXR_TERM)
  {
    loadcount--;
#ifdef DEBUG
    fprintf(fp,"Checking for unload - loadcount = %d\n",loadcount);
#endif
    if (loadcount==0 && ep != NULL)
    {
#ifdef DEBUG
      fprintf(fp,"Unloading %s\n",MQSECCE);
#endif
    /*************************************************************************/
    /* Normally, I'd unload the module at this point, when it is             */
    /* no longer needed. But for some reason, it was causing a trap when     */
    /* the wrapper was also being unloaded. So I've removed the line for now.*/
    /*************************************************************************/
    /*terminateAndUnload((void *)ep);*/

      ep = NULL;
    }
  }

  /***************************************************************************/
  /* And finally, whatever else happened, release the global lock.           */
  /***************************************************************************/
  pthread_mutex_unlock(&m);

  if ((loadcount==0) && (pParms->ExitReason == MQXR_TERM))
  {
    if (mutex_avail)
    {
      pthread_mutex_destroy(&m);
      mutex_avail = 0;
    }
  }
#ifdef DEBUG
  fclose(fp);
#endif
  return;
}

static void onceonly(void)
{
  /***************************************************************************/
  /* This function is called only once, to initialise the global mutex.      */
  /* The pthreads definition assures that this is actually invoked only once.*/
  /***************************************************************************/
  pthread_mutex_init(&m,NULL);
  mutex_avail = 1;
  return;
}
